#include "drawtree.h"

#define HEIGHT_LEVEL 30

#define MAX(X,Y) ((X)<(Y)?(Y):(X))

int bst_size(bst_t *tree)
{
    return tree->size;
}

int bst_height_helper(bstnode_t *node)
{
    if (node == NULL) return -1;
    else return 1 + MAX(bst_height_helper(node->left), bst_height_helper(node->right));
}

int bst_height(bst_t *tree)
{
    return bst_height_helper(tree->root);
}

/*
 * Determine the number of pixels needed to draw a label.
 * (assumes data_t is an int).
 */
int pixel_width_label(bstnode_t *node)
{
    if (node->data < 10) {
        return 10;
    } else {
        return 5 + 5*(int)(ceil(log10((int)node->data)));
    }
}

/*
 * Recursively determine the number of pixels needed to represent an entire
 * subtree.
 */
int pixel_width(bstnode_t *node, int width_scale)
{
    if (node == NULL) {
        return 0;
    } else {
        return (pixel_width_label(node) +
                pixel_width(node->left, width_scale) +
                pixel_width(node->right, width_scale));
    }
}

/*
 * Determine the number of pixels needed to represent this node and its entire
 * left subtree.
 */
int pixel_width_left(bstnode_t *node, int width_scale)
{
    if (node == NULL) {
        return 0;
    } else {
        return (pixel_width_label(node) +
                pixel_width(node->left, width_scale));
    }
}

/*
 * Determine the number of pixels needed to represent this node and its entire
 * right subtree.
 */
int pixel_width_right(bstnode_t *node, int width_scale)
{
    if (node == NULL) {
        return 0;
    } else {
        return (pixel_width_label(node) +
                pixel_width(node->right, width_scale));
    }
}

#define HEIGHT_LEVEL 30

/*
 * Recursive helper for turtle_draw_bst.
 */
void turtle_draw_bstnode(bstnode_t *node, int text_offset, int width_scale, dynarray_t *highlight_values)
{
    if (node == NULL) return;

    // save current position
    double cx = turtle_get_x();
    double cy = turtle_get_y();

    // draw label
    turtle_pen_up();
    turtle_set_heading(90);
    turtle_forward(10);
    turtle_set_heading(0);
    turtle_forward(text_offset-2);
    turtle_set_pen_color(0, 0, 0);
    turtle_draw_int(node->data);

    // draw node dot
    turtle_goto(cx, cy);
    turtle_pen_down();
    turtle_set_pen_color(0, 0, 255);
    if (highlight_values != NULL && dynarray_contains(highlight_values, node->data)) {
        turtle_set_fill_color(255, 0, 0);
    } else {
        turtle_set_fill_color(0, 0, 255);
    }
    turtle_fill_circle(cx, cy, 5);

    // recursive call (left side -> negative text position offset)
    if (node->left != NULL) {
        turtle_goto(cx - pixel_width_right(node->left, width_scale),
                    cy - (double)HEIGHT_LEVEL);
        turtle_draw_bstnode(node->left, -5, width_scale, highlight_values);
        turtle_pen_up();
        turtle_goto(cx, cy);
        turtle_pen_down();
    }

    // recursive call (right side -> positive text position offset)
    if (node->right != NULL) {
        turtle_goto(cx + pixel_width_left(node->right, width_scale),
                    cy - (double)HEIGHT_LEVEL);
        turtle_draw_bstnode(node->right, 5, width_scale, highlight_values);
        turtle_pen_up();
        turtle_goto(cx, cy);
        turtle_pen_down();
    }
}

/*
 * Draw a binary tree using turtle graphics.
 */
void turtle_draw_bst(bst_t *tree, dynarray_t *highlight_values)
{
    if (tree->root == NULL) return;

    // variable scaling for the tree width
    // (should get thinner as the tree grows)
    int width_scale = (int)(75.0 / MAX(1.0, sqrt(bst_size(tree))));

    // calculate and configure diagram size
    int screen_width = pixel_width(tree->root, width_scale);
    int right_subtree_width = 0;
    if (tree->root != NULL && tree->root->right != NULL) {
        right_subtree_width = pixel_width(tree->root->right, width_scale);
    }
    turtle_init(screen_width + 50, (bst_height(tree)+1) * HEIGHT_LEVEL + 50);

    // go to starting point
    turtle_pen_up();
    turtle_goto((int)(0.5 * (double)screen_width - ((double)right_subtree_width + 10.0)),
                (int)((double)bst_height(tree) * (double)HEIGHT_LEVEL / 2.0));
    turtle_pen_down();

    // recursively draw tree, starting at the root
    turtle_draw_bstnode(tree->root, 0, width_scale, highlight_values);

    // save output diagram and clean up
    turtle_save_bmp("output.bmp");
    turtle_cleanup();
}
